/*

Copyright (C) 2023 Dirk Schuetz <dirk.schuetz@durham.ac.uk>

This file is part of KnotJob.

KnotJob is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

KnotJob is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org.licenses/>.

 */

package knotjob;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import javax.swing.JButton;
import javax.swing.JPanel;
import knotjob.dialogs.CalculationDialog;
import knotjob.dialogs.CalculationDialogWrap;
import knotjob.dialogs.CharDialog;
import knotjob.dialogs.KhovDialog;
import knotjob.dialogs.SquareDialog;
import knotjob.dialogs.StableDialog;
import knotjob.homology.evenkhov.EvenKhovCalculator;
import knotjob.homology.evenkhov.sinv.GradedCalculator;
import knotjob.homology.evenkhov.sinv.SInvariantCalculator;
import knotjob.homology.evenkhov.sinv.SpectralCalculator;
import knotjob.homology.evenkhov.sinv.SqOneCalculator;
import knotjob.homology.evenkhov.sinv.UnivKhovCalculator;
import knotjob.homology.oddkhov.OddKhovCalculator;
import knotjob.homology.oddkhov.sinv.SqOneOddCalculator;
import knotjob.homology.oddkhov.unified.UnifiedKhovCalculator;
import knotjob.homology.oddkhov.unified.sinv.CompleteSCalculator;
import knotjob.homology.oddkhov.unified.sinv.SqOneSumCalculator;
import knotjob.homology.oddkhov.unified.sinv.UnifiedSCalculator;
import knotjob.invariants.SignatureCalculator;
import knotjob.links.LinkData;
import knotjob.polynomial.alexander.AlexanderCalculator;
import knotjob.polynomial.jones.JonesCalculator;
import knotjob.stabletype.StableEvenCalculator;
import knotjob.stabletype.StableOddCalculator;
import knotjob.stabletype.sinv.SqTwoCalculator;
import knotjob.stabletype.sinv.SqTwoOddCalculator;

/**
 *
 * @author Dirk
 */
public class KnotButtons {

    private final Options options;
    private final Knobster knobster;
    
    public KnotButtons(Options optns, Knobster knbstr) {
        options = optns;
        knobster = knbstr;
    }
    
    void addButtonsTo(JPanel panelButtons) {
        JButton calcSInvariant = new JButton("s-Invariants");
        calcSInvariant.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                sInvAction();
            }
        });
        JButton lipSarInvariant = new JButton("Lipshitz-Sarkar Invariants");
        lipSarInvariant.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                lipSarAction();
            }
        });
        JButton calcModKhovHom = new JButton("Modular Khovanov Cohomology");
        calcModKhovHom.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                khovhomAction(true);
            }
        });
        JButton calcModOddHom = new JButton("Modular Odd Khovanov Homology");
        calcModOddHom.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                oddhomAction(true);
            }
        });
        JButton calcChaKhovHom = new JButton("Char 0 Khovanov Cohomology");
        calcChaKhovHom.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                khovhomAction(false);
            }
        });
        JButton calcChaOddHom = new JButton("Char 0 Odd Khovanov Homology");
        calcChaOddHom.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                oddhomAction(false);
            }
        });
        JButton gradedInv = new JButton("Graded s-Invariants");
        gradedInv.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                gradedAction();
            }
        });
        JButton bLTSS = new JButton("BLT-Spectral Sequence");
        bLTSS.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                spectralAction();
            }
        });
        JButton stableType = new JButton("Stable Homotopy Type");
        stableType.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                stableAction();
            }
        });
        JButton signature = new JButton("Signature");
        signature.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                signAction();
            }
        });
        JButton jonesPolynomial = new JButton("Jones Polynomial");
        jonesPolynomial.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                jonesAction();
            }
        });
        JButton alexPolynomial = new JButton("Alexander Polynomial");
        alexPolynomial.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent ae) {
                alexAction();
            }
        });
        JPanel calcSPanel = buttonPanel(calcSInvariant);
        JPanel lipSarPanel = buttonPanel(lipSarInvariant);
        JPanel modevKhPanel = buttonPanel(calcModKhovHom);
        JPanel charevKhPanel = buttonPanel(calcChaKhovHom);
        JPanel mododKhPanel = buttonPanel(calcModOddHom);
        JPanel charodKhPanel = buttonPanel(calcChaOddHom);
        JPanel gradedPanel = buttonPanel(gradedInv);
        JPanel bLTSSPanel = buttonPanel(bLTSS);
        JPanel signaturePanel = buttonPanel(signature);
        JPanel jonesPanel = buttonPanel(jonesPolynomial);
        JPanel alexPanel = buttonPanel(alexPolynomial);
        JPanel stablePanel = buttonPanel(stableType);
        panelButtons.setLayout(new GridLayout(6,3));
        for (int i = 0; i < 3; i++) panelButtons.add(buttonPanel(null));
        panelButtons.add(calcSPanel);
        panelButtons.add(charevKhPanel);
        panelButtons.add(charodKhPanel);
        panelButtons.add(lipSarPanel);
        panelButtons.add(modevKhPanel);
        panelButtons.add(mododKhPanel);
        panelButtons.add(gradedPanel);
        panelButtons.add(bLTSSPanel);
        panelButtons.add(stablePanel);
        panelButtons.add(jonesPanel);
        panelButtons.add(alexPanel);
        panelButtons.add(signaturePanel);
        for (int i = 0; i < 3; i++) panelButtons.add(buttonPanel(null));
    }
    
    private JPanel buttonPanel(JButton theButton) {
        JPanel thePanel = new JPanel();
        thePanel.setBackground(options.getColor());
        if (theButton != null) thePanel.add(theButton);
        return thePanel;
    }
    
    protected void sInvAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateSInvariant(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void lipSarAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateLipSar(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void gradedAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateGraded(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void stableAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateStableType(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void spectralAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateSpectralSequence(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void khovhomAction(boolean modular) {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateKhovHom(chosenLinks(linkList, knobster.choices), modular);
            }
        }
    }
    
    protected void oddhomAction(boolean modular) {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateOddKhovHom(chosenLinks(linkList, knobster.choices), modular);
            }
        }
    }
    
    protected void signAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateSignature(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void jonesAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateJones(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    protected void alexAction() {
        if (knobster.choices != null) {
            if (knobster.choices.length > 0) {
                ArrayList<LinkData> linkList = knobster.allLinks;
                if (knobster.filtered) linkList = knobster.filteredLinks;
                calculateAlexander(chosenLinks(linkList, knobster.choices));
            }
        }
    }
    
    private ArrayList<LinkData> chosenLinks(ArrayList<LinkData> linkList, int[] choices) {
        ArrayList<LinkData> theList = new ArrayList<LinkData>(choices.length);
        for (int c : choices) theList.add(linkList.get(c));
        return theList;
    }
    
    protected void calculateUnivKhov(ArrayList<LinkData> linkList) {
        String title = "Universal Khovanov";
        KhovDialog diagFrame = new KhovDialog(knobster, true, title, false);
        int val = (int) diagFrame.getValue();
        if (val < 0) return;
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, title, lev);
        UnivKhovCalculator calculator = new UnivKhovCalculator(linkList, val, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    protected void calculateUnifiedKhov(ArrayList<LinkData> linkList) {
        String title = "Unified Khovanov";
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, title, lev);
        UnifiedKhovCalculator calculator = new UnifiedKhovCalculator(linkList, 1, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateKhovHom(ArrayList<LinkData> linkList, boolean mod) {
        KhovDialog diagFrame = new KhovDialog(knobster, mod);
        long val = diagFrame.getValue();
        diagFrame.dispose();
        if (diagFrame.isOkay()) calculateKnotHom(linkList,val);
    }
    
    private void calculateOddKhovHom(ArrayList<LinkData> linkList, boolean mod) {
        KhovDialog diagFrame = new KhovDialog(knobster, mod);
        diagFrame.setTitle("Odd Khovanov Homology");
        long val = diagFrame.getValue();
        diagFrame.dispose();
        if (diagFrame.isOkay()) calculateOddKnotHom(linkList,val);
    }
    
    private void calculateKnotHom(ArrayList<LinkData> linkList, long val) {
        String khovTitle = "Integral Khovanov Cohomology";
        if (val == 1) khovTitle = "Rational Khovanov Cohomology";
        if (val < 0) khovTitle = "Local Khovanov Cohomology";
        if (val > 1) khovTitle = "Khovanov Cohomology mod "+val;
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, khovTitle, lev);
        EvenKhovCalculator calculator = new EvenKhovCalculator(linkList, val, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateOddKnotHom(ArrayList<LinkData> linkList, long val) {
        String khovTitle = "Integral Odd Khovanov Homology";
        if (val == 1) khovTitle = "Rational Odd Khovanov Homology";
        if (val < 0) khovTitle = "Local Odd Khovanov Homology";
        if (val > 1) khovTitle = "Odd Khovanov Homology mod "+val;
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, khovTitle, lev);
        OddKhovCalculator calculator = new OddKhovCalculator(linkList, val, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    /*protected void calculateSlThreeHom(ArrayList<LinkData> linkList) {
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate sl_3-Homology", 2);
        SlThreeHomCalculator calculator = new SlThreeHomCalculator(linkList, 0, options,
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }// */
    
    private void calculateSInvariant(ArrayList<LinkData> linkList) {
        CharDialog charFrame = new CharDialog(knobster, "s-Invariant", true, options, false);
        int p = charFrame.getChar();
        charFrame.dispose();
        if (p<0) return;
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate s-Invariant mod "+p, 
                lev);
        SInvariantCalculator calculator = new SInvariantCalculator(linkList, p, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateGraded(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate graded s-Invariant", lev);
        GradedCalculator calculator = new GradedCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateStableType(ArrayList<LinkData> linkList) {
        StableDialog staFrame = new StableDialog(knobster, "Stable Homotopy Types", true);
        int par = staFrame.getParity();
        staFrame.dispose();
        String title = "Calculate Stable Type";
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, title, lev);
        if (par < 0) return;
        if (par == 0) {
            StableEvenCalculator calculator = new StableEvenCalculator(linkList, options, 
                    new CalculationDialogWrap(frame, options.getGirthInfo() >0));
            calculator.start();
        }
        if (par >= 1) {
            StableOddCalculator calculator = new StableOddCalculator(linkList, options, 
                    new CalculationDialogWrap(frame, options.getGirthInfo() >0), par);
            calculator.start();
        }
        frame.setUpStuff();
    }
    
    private void calculateSpectralSequence(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CharDialog charFrame = new CharDialog(knobster, "BLT-Spectral Sequence", true, options, true);
        int p = charFrame.getChar();
        if (p < 0) return;
        boolean red = charFrame.getReduced();
        charFrame.dispose();
        String title = "Calculate BLTSS mod "+p;
        if (red) title = title + " reduced";
        CalculationDialog frame = new CalculationDialog(knobster, title, lev);
        SpectralCalculator calculator = new SpectralCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0), p, red);
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateLipSar(ArrayList<LinkData> linkList) {
        SquareDialog squaFrame = new SquareDialog(knobster, "Lipshitz-Sarkar Invariants", true);
        int dig = squaFrame.getSquare();
        squaFrame.dispose();
        if (dig == 1) calculateLipSarEvenSqOne(linkList);
        if (dig == 3) calculateBLSOdd(linkList);
        if (dig == 4) calculateLipSarSumSqOne(linkList);
        if (dig == -1) calculateLipSarOddSqOne(linkList);
        if (dig == 2) calculateLipSarEvenSqTwo(linkList);
        if (dig == -2) calculateLipSarOddSqTwo(linkList, dig+2);
        if (dig == -3) calculateLipSarOddSqTwo(linkList, dig+4);
        if (dig == -4) calculateUnifiedInv(linkList);
        if (dig == -5) calculateCompleteSInv(linkList);
    }
    
    protected void calculateCompleteSInv(ArrayList<LinkData> linkList) {
        String title = "Complete s-Invariant";
        //ReducedDialog dialog = new ReducedDialog(knobster, title, true);
        //dialog.showDialog();
        //int dec = dialog.getDecision();
        //if (dec == 0) return;
        //boolean red = true;
        title = title+" reduced";
        //else title = title+" unreduced";
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, title, lev);
        CompleteSCalculator calculator = new CompleteSCalculator(linkList, options,
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0), true);
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateUnifiedInv(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate Unified s-Invariant", lev);
        UnifiedSCalculator calculator = new UnifiedSCalculator(linkList, options,
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateLipSarSumSqOne(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate Sq^1 sum s-Invariant", lev);
        SqOneSumCalculator calculator = new SqOneSumCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateLipSarOddSqOne(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate odd Sq^1 s-Invariant", lev);
        SqOneOddCalculator calculator = new SqOneOddCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0), 4);
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateBLSOdd(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate odd BLS s-Invariant", lev);
        SqOneOddCalculator calculator = new SqOneOddCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0), 32768);
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateLipSarOddSqTwo(ArrayList<LinkData> linkList, int eps) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate odd Sq^2 ("+
                (char)949+"="+eps+") s-Invariant", lev);
        SqTwoOddCalculator calculator = new SqTwoOddCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0), eps);
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateLipSarEvenSqOne(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate even Sq^1 s-Invariant", lev);
        SqOneCalculator calculator = new SqOneCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateLipSarEvenSqTwo(ArrayList<LinkData> linkList) {
        int lev = 3;
        if (options.getGirthInfo() == 2) lev = 4;
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate even Sq^2 s-Invariant", lev);
        SqTwoCalculator calculator = new SqTwoCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0), 
                options.getStableBound(), options.getCombAtt(), options.getCombine());
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateSignature(ArrayList<LinkData> linkList) {
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate Signature", 1);
        SignatureCalculator calculator = new SignatureCalculator(linkList, options,
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateAlexander(ArrayList<LinkData> linkList) {
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate Alexander Polynomial", 2);
        AlexanderCalculator calculator = new AlexanderCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    private void calculateJones(ArrayList<LinkData> linkList) {
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate Jones Polynomial", 2);
        JonesCalculator calculator = new JonesCalculator(linkList, options, 
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }
    
    /*protected void calculateSlThreePol(ArrayList<LinkData> linkList) {
        CalculationDialog frame = new CalculationDialog(knobster, "Calculate sl_3-Polynomial", 2);
        SlThreePolCalculator calculator = new SlThreePolCalculator(linkList, options,
                new CalculationDialogWrap(frame, options.getGirthInfo() > 0));
        calculator.start();
        frame.setUpStuff();
    }// */
}
