/*

Copyright (C) 2019-20 Dirk Schuetz <dirk.schuetz@durham.ac.uk>

This file is part of KnotJob.

KnotJob is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

KnotJob is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org.licenses/>.

 */

package knotjob.filters;

import knotjob.links.LinkData;

/**
 *
 * @author Dirk
 */
public class SInvFilter implements Filter {

    private final boolean nonConstS;
    private final boolean nonConstSqOne;
    private final boolean nonConstSqTwo;
    private final boolean nonStdtGrd;
    private final boolean even;
    private final boolean bddBelow;
    private final boolean bddAbove;
    private final int sqTwoCase; // 0 : odd/eps = 0; 1 : odd/eps = 1; 2 : even
    private final int lowerBound;
    private final int upperBound;
    private String name;
    
    public SInvFilter(String nm, boolean ncS, boolean ncSq, boolean ev, boolean bddb, boolean bdda, 
            int lb, int ub) {
        nonConstS = ncS;
        nonConstSqOne = ncSq;
        nonStdtGrd = false;
        even = ev;
        bddBelow = bddb;
        bddAbove = bdda;
        lowerBound = lb;
        upperBound = ub;//upperBound == 1 means we're looking for BLSOdd
        name = nm;
        sqTwoCase = 0;
        nonConstSqTwo = false;
    }
    
    public SInvFilter(String nm, int stcase) {
        nonStdtGrd = false;
        nonConstS = false;
        nonConstSqOne = false;
        nonConstSqTwo = true;
        even = false;
        bddBelow = false;
        bddAbove = false;
        sqTwoCase = stcase;
        lowerBound = 0;
        upperBound = 0;
        name = nm;
    }
    
    public SInvFilter(String nm, boolean nstd) {
        nonStdtGrd = nstd;
        nonConstS = false;
        nonConstSqOne = false;
        nonConstSqTwo = false;
        even = false;
        bddBelow = false;
        bddAbove = false;
        sqTwoCase = 0;
        lowerBound = 0;
        upperBound = 0;
        name = nm;
    }
    
    @Override
    public boolean linkIsFiltered(LinkData link) {
        boolean okay = false;
        if (nonStdtGrd) return linkHasNonStdtGrd(link);
        if (nonConstS) okay = linkHasNonConstS(link);
        if (nonConstSqOne) okay = linkHasNonConstSqOne(link);
        if (nonConstSqTwo) return linkHasNonConstSqTwo(link);
        if (!nonConstS && !nonConstSqOne ) okay = sInvInBounds(link);
        return okay;
    }
    
    private boolean sInvInBounds(LinkData link) {
        boolean okay = false;
        int[][] sinvs = link.sInvariants();
        int i = 0;
        while (!okay & i < sinvs.length) {
            int s = sinvs[i][1];
            okay = true;
            if (bddBelow && s < lowerBound) okay = false;
            if (bddAbove && s > upperBound) okay = false;
            i++;
        }
        return okay;
    }
    
    private boolean linkHasNonStdtGrd(LinkData link) {
        if (link.grsinv == null) return false;
        return link.grsinv.contains("(");
    }
    
    private boolean linkHasNonConstSqOne(LinkData link) {
        boolean okay = false;
        int[] sqOne = link.getSqOne(even);
        if (upperBound == 1) sqOne = link.getBLSOdd();
        if (upperBound == 2) sqOne = link.getSqOneSum();
        if (upperBound == 4) sqOne = link.getCmif();
        if (sqOne != null) {
            int c = sqOne[0];
            okay = false;
            if (sqOne[1] != c | sqOne[2] != c | sqOne[3] != c) okay = true;
        }
        return okay;
    }
    
    private boolean linkHasNonConstSqTwo(LinkData link) {
        boolean okay = false;
        int[] sqTwo = link.getSqTwo(sqTwoCase);
        if (sqTwo != null) {
            int c = sqTwo[0];
            okay = false;
            if (sqTwo[1] != c | sqTwo[2] != c | sqTwo[3] != c) okay = true;
        }
        return okay;
    }
    
    private boolean linkHasNonConstS(LinkData link) {
        boolean okay = false;
        int[][] sinv = link.sInvariants();
        if (sinv.length > 0) {
            int i = 1;
            okay = false;
            int c = sinv[0][1];
            while (i < sinv.length && !okay) {
                if (sinv[i][1] != c) okay = true;
                else i++;
            }
        }
        return okay;
    }
    
    @Override
    public String getName() {
        return name;
    }

    @Override
    public void setName(String nm) {
        name = nm;
    }
    
}
