/*

Copyright (C) 2021 Dirk Schuetz <dirk.schuetz@durham.ac.uk>

This file is part of KnotJob.

KnotJob is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

KnotJob is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org.licenses/>.

 */

package knotjob.homology.evenkhov.sinv;

import java.util.ArrayList;
import knotjob.AbortInfo;
import knotjob.Options;
import knotjob.dialogs.DialogWrap;
import knotjob.links.Link;
import knotjob.links.LinkData;
import knotjob.links.Reidemeister;
import knotjob.rings.Ring;

/**
 *
 * @author Dirk
 * @param <R>
 */
public class SpectralInvariant<R extends Ring<R>> {
    
    private final Link theLink;
    private final LinkData theData;
    private final DialogWrap frame;
    private final AbortInfo abInf;
    private final boolean reduced;
    private final boolean lee;
    private final ArrayList<String> allInfo;
    private final ArrayList<Integer> blocks;
    private final R unit;
    private final int[] girth;
    private final boolean highDetail;
    
    public SpectralInvariant(LinkData tlnk, R unt, DialogWrap frm, Options optns, boolean red,
            boolean le) {
        Link lnk = Reidemeister.freeOfOne(tlnk.chosenLink().breakUp().girthMinimize());
        if (lnk.crossingLength() == 1) theLink = lnk.breakUp();
        else theLink = lnk;
        theData = tlnk;
        unit = unt;
        reduced = red;
        lee = le;
        allInfo = new ArrayList<String>();
        blocks = new ArrayList<Integer>();
        girth = theLink.totalGirthArray();
        frame = frm;
        abInf = frm.getAbortInfo();
        highDetail = optns.getGirthInfo() == 2;
    }
    
    public ArrayList<String> getInfo() {
        return allInfo;
    }
    
    public ArrayList<Integer> getBlocks() {
        return blocks;
    }
    
    public void calculate() {
        int[] wrt = theLink.crossingSigns();
        int hstart = -wrt[1];
        int qstart = wrt[0]+2*hstart;
        calculateSpecSeq(hstart, qstart);
    }
    
    private void calculateSpecSeq(int hs, int qs) {
        SInvComplex<R> theComplex = getComplex(hs, qs);
        if (abInf.isAborted()) return;
        int bn = 2;
        if (lee) bn = 4;
        int qJump = bn;
        int shift = theData.bltShift(reduced);
        blocks.add(1);
        ArrayList<String> qCohs = theComplex.homologyInfo();
        copyOver(qCohs);
        blocks.add(qCohs.size());
        int eCounter = 2;
        while (theComplex.containsBoundaries()) {
            theComplex.cancelGenerators(qJump);
            if (abInf.isAborted()) return;
            qCohs = theComplex.homologyInfo();
            blocks.add(allInfo.size()+1);
            copyOver(qCohs);
            blocks.add(allInfo.size());
            qJump = qJump + bn;
            eCounter++;
        }
    }
    
    private void copyOver(ArrayList<String> strings) {
        for (String dtr : strings) allInfo.add(dtr);
    }
    
    private SInvComplex<R> getComplex(int hs, int qs) {
        if (theLink.crossingLength() == 0) return new SInvComplex<R>(theLink.unComponents()-1, unit,
                !reduced, reduced, abInf, null);
        int tsum = totalSum(theLink.getCrossings());
        SInvComplex<R> theComplex = new SInvComplex<R>(theLink.getCross(0), theLink.getPath(0),
                hs, qs, false, true, true, false, unit, frame, abInf);
        int u = 1;
        while (u < theLink.crossingLength()-1) {
            boolean orient = (theComplex.negContains(theLink.getPath(u, 0))| theComplex.negContains(theLink.getPath(u, 2))|
                theComplex.posContains(theLink.getPath(u,1)) | theComplex.posContains(theLink.getPath(u,3)));
            SInvComplex<R> nextComplex = new SInvComplex<R>(theLink.getCross(u), theLink.getPath(u),
                    0, 0, orient, true, true, false, unit, null, null);
            frame.setLabelRight(String.valueOf(u+1)+"/"+String.valueOf(theLink.crossingLength()), 0, false);
            theComplex.modifyComplex(nextComplex,0,girthInfo(u), highDetail);
            if (theLink.getCross(u) < 0) tsum = tsum - theLink.getCross(u);
            else tsum = tsum + theLink.getCross(u);
            u++;
        }
        theComplex = lastComplex(theComplex, u);
        //theComplex.boundaryCheck();
        return theComplex;
    }
    
    private SInvComplex<R> lastComplex(SInvComplex<R> theComplex, int u) {
        boolean orient = (theComplex.negContains(theLink.getPath(u, 0))| theComplex.negContains(theLink.getPath(u, 2))|
            theComplex.posContains(theLink.getPath(u,1)) | theComplex.posContains(theLink.getPath(u,3)));
        GradedComplex<R> nextComplex = new GradedComplex<R>(theLink.getCross(u), theLink.getPath(u),
                0, 0, orient, false, true, false, unit, null, null);
        frame.setLabelRight(String.valueOf(u+1)+"/"+String.valueOf(theLink.crossingLength()), 0, false);
        int c = 0;
        if (reduced) c = 1+theLink.getPath(u, theLink.basepoint());
        theComplex.modifyComplex(nextComplex, c, "0", highDetail);
        return theComplex;
    }
    
    private String girthInfo(int u) {
        String info = String.valueOf(girth[u]);
        if (!highDetail) return info;
        if (u < girth.length - 1) info = info+" ("+girth[u+1];
        else return info;
        for (int i = 1; i < 3; i++) {
            if (u < girth.length - i - 1) info = info+", "+girth[u+1+i];
        }
        info = info+")";
        return info;
    }

    private int totalSum(int[] crossings) {
        int tsum = 0;
        for (int r = 1; r < crossings.length; r++) {
            if (crossings[r] < 0) tsum = tsum + crossings[r];
            else tsum = tsum - crossings[r];
        }
        tsum--;
        return tsum;
    }
    
}
