/*

Copyright (C) 2021-22 Dirk Schuetz <dirk.schuetz@durham.ac.uk>

This file is part of KnotJob.

KnotJob is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

KnotJob is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org.licenses/>.

 */

package knotjob.links;

import java.util.ArrayList;

/**
 *
 * @author Dirk
 */
public class Reidemeister {
    
    public static boolean allFour(Link link) {
        boolean allfour = true;
        int i = 0;
        while (allfour && i < link.crossingLength()) {
            ArrayList<Integer> helper = allOfCrossing(link.getPath(i));
            allfour = (helper.size() == 4);
            i++;
        }
        return allfour;
    }
    
    public static Link freeOfOne(Link link) {
        if (allFour(link)) return link;
        int i = firstNotFour(link);
        while (i >= 0) {
            int cr = link.getCross(i);
            if (Math.abs(cr) != 1) link = splitUp(link, i);
            else {
                ArrayList<Integer> helper = allOfCrossing(link.getPath(i));
                if (helper.isEmpty()) { // it is an unknot component
                    link = removeUnknot(link, i);
                }
                else {
                    link = removeROne(link, i);
                }
            }
            i = firstNotFour(link);
        }
        return link;
    }

    private static int firstNotFour(Link link) {
        int i = 0;
        boolean found = false;
        while (!found && i < link.crossingLength()) {
            if (allOfCrossing(link.getPath(i)).size() < 4) found = true;
            else i++;
        }
        if (found) return i;
        return -1;
    }

    private static ArrayList<Integer> allOfCrossing(int[] path) {
        ArrayList<Integer> helper = new ArrayList<Integer>(4);
        for (int u : path) {
            if (!helper.contains(u)) helper.add(u);
            else helper.remove((Integer) u);
        }
        return helper;
    }

    private static Link splitUp(Link link, int i) {
        int n = link.getCross(i);
        int m = Math.abs(n);
        int[] crossings = new int[link.crossingLength()+m-1];
        int[][] paths = new int[link.crossingLength()+m-1][4];
        for (int j = 0; j < link.crossingLength(); j++) {
            int k = j;
            if (k > i) k = k + m - 1;
            crossings[k] = link.getCross(j);
            for (int l = 0; l < 4; l++) paths[k][l] = link.getPath(j, l);
        }
        int om = 2 * link.crossingLength();
        for (int j = 0; j < m - 1; j++) {
            paths[i+j][1] = om + 2 * j + 1;
            paths[i+j][2] = om + 2 * j + 2;
            paths[i+j+1][0] = om + 2 * j + 1;
            paths[i+j+1][3] = om + 2 * j + 2;
            crossings[i+j] = n/m;
        }
        crossings[i+m-1] = n/m;
        paths[i+m-1][1] = link.getPath(i, 1);
        paths[i+m-1][2] = link.getPath(i, 2);
        ArrayList<int[]> ors = new ArrayList<int[]>();
        for (int j = 0; j < link.relComponents(); j++) {
            int[] or = link.orientation(j);
            if (or[0] > i) {
                int[] nor = new int[2];
                if (or[0] > i) nor[0] = or[0]+m-1;
                else nor[0] = or[0];
                nor[1] = or[1];
                ors.add(nor);
            }
            else ors.add(or);
        }
        return new Link(crossings, paths, ors, link.unComponents());
    }

    private static Link removeUnknot(Link link, int i) {
        int a = link.getPath(i, 0);
        int b = link.getPath(i, 2);
        int[][] paths = new int[link.crossingLength()-1][4];
        int[] crossings = new int[link.crossingLength()-1];
        for (int j = 0; j < link.crossingLength(); j++) {
            if (j != i) {
                int k = j;
                if (j > i) k--;
                crossings[k] = link.getCross(j);
                for (int l = 0; l < 4; l++) {
                    int path = link.getPath(j, l);
                    if (path >= a) path--;
                    if (path >= b) path--;
                    paths[k][l] = path;
                }
            }
            
        }
        ArrayList<int[]> ors = new ArrayList<int[]>();
        for (int j = 0; j < link.relComponents(); j++) {
            int[] or = link.orientation(j);
            if (or[0] != i) {
                int[] nor = new int[2];
                if (or[0] > i) nor[0] = or[0]-1;
                else nor[0] = or[0];
                nor[1] = or[1];
                ors.add(nor);
            }
        }
        return new Link(crossings, paths, ors, link.unComponents()+1);
    }

    private static Link removeROne(Link link, int i) {
        int a = link.getPath(i, 0);
        int b = link.getPath(i, 2);
        int c = link.getPath(i, 1);
        int d = link.getPath(i, 3);
        int[][] paths = new int[link.crossingLength()-1][4];
        int[] crossings = new int[link.crossingLength()-1];
        for (int k = 0; k < link.crossingLength()-1; k++) {
            int l = k;
            if (l >= i) l++;
            crossings[k] = link.getCross(l);
            for (int m = 0; m < 4; m++) {
                paths[k][m] = link.getPath(l, m);
            }
        }
        if (a == c || a == d) {
            if (a == c) {
                c = d;
            }
        }
        else {
            if (b == c) {
                b = a;
                a = c;
                c = d;
            }
            else { // b == d
                b = a;
                a = d;
            }
        } // we now have that a = d, while b and c are different, and are part of the rest of the link
        ArrayList<int[]> ors = new ArrayList<int[]>();
        for (int j = 0; j < link.relComponents(); j++) {
            int[] or = link.orientation(j);
            if (or[0] != i) {
                int[] nor = new int[2];
                if (or[0] > i) nor[0] = or[0]-1;
                else nor[0] = or[0];
                nor[1] = or[1];
                ors.add(nor);
            }
            else {
                ArrayList<ArrayList<Integer>> comps = link.getComponents();
                int k = componentWith(a, comps);
                boolean btoc = getOrder(comps.get(k), a, b, c);
                int[] nor;
                if (btoc) nor = posOf(c, paths);
                else nor = posOf(b, paths);
                ors.add(nor);
            }
        }
        if (b > c) {
            d = c;
            c = b;
            b = d;
        } // we now have b < c
        for (int m = 0; m < paths.length; m++) {
            for (int n = 0; n < 4; n++) {
                int p = paths[m][n];
                if (p == c) p = b;
                if (p > a) p--;
                if (p >= c) p--;
                paths[m][n] = p;
            }
        }
        return new Link(crossings, paths, ors, link.unComponents());
    }

    private static int componentWith(int a, ArrayList<ArrayList<Integer>> comps) {
        boolean found = false;
        int i = 0;
        while (!found) {
            if (comps.get(i).contains(a)) found = true;
            else i++;
        }
        return i;
    }

    private static boolean getOrder(ArrayList<Integer> comp, int a, int b, int c) {
        int u = comp.indexOf(a);
        int m = comp.size();
        if (u == 0) {
            comp.add(0, comp.get(m-1));
            comp.remove(m);
        }
        if (u == m-1) {
            comp.add(comp.get(0));
            comp.remove(0);
        } // we now have the order b a c or c a b with no cyclic overlap
        int v = comp.indexOf(b);
        u = comp.indexOf(c);
        return (v < u); // return (pos b is less than pos c)
    }

    private static int[] posOf(int c, int[][] paths) {
        boolean found = false;
        int i = 0;
        int j = 0;
        while (!found) {
            j = 0;
            while (!found && j < 4) {
                if (paths[i][j] == c) found = true;
                else j++;
            }
            if (!found) i++;
        }
        return new int[] {i, j};
    }
    
    public static Link combineCrossings(Link link) {
        int i = 0;
        boolean sequence = false;
        int s1 = 0;
        int s2 = 0;
        int q1 = 0;
        int q2 = 0;//q's are position, s's are values
        int[] crossings = link.cloneCrossings();
        int[][] paths = link.clonePaths();
        ArrayList<int[]> ors = link.cloneOr();
        while (i < crossings.length-1) {
            if (sequence) {
                int u = i+1;
                boolean found = false;
                while (u < crossings.length && !found) {
                    if (contains(paths[u], s1, s2, crossings[u])) {
                        found = true;
                        boolean fnd = false;
                        int y = 0;
                        while (!fnd) {
                            if (paths[u][y] == s2) fnd = true;
                            else y++;
                        }
                        int t1 = (y+1)%4;
                        int p1 = (y+2)%4;
                        int p2 = (y+3)%4;
                        for (int[] or : ors) {
                            if (or[0] == i) {
                                if (or[1] == (q2+2)%4 || or[1] == (q2+3)%4) {
                                    if ((Math.abs(crossings[u]))%2 != 0) {
                                        if (or[1] %2 != 0) or[1] = (or[1] + 1)%4;
                                        else or[1] = (or[1] + 3)%4;
                                    }
                                } 
                            }
                            if (or[0] == u) {
                                or[0] = i;
                                if (or[1] == y || or[1] == t1) {
                                    if (Math.abs(crossings[i])%2 != 0) or[1] = q2;
                                    else or[1] = (q2+1)%4;
                                }
                                else if (y % 2 == 0) or[1] = (or[1]+1)%4;
                            }
                        }
                        int[] newcrossings = new int[crossings.length-1];
                        System.arraycopy(crossings, 0, newcrossings, 0, i);
                        if (t1%2 != 0) newcrossings[i] = crossings[i]-crossings[u];
                        else newcrossings[i] = crossings[i]+crossings[u];
                        System.arraycopy(crossings, i+1, newcrossings, i+1, u - (i+1));
                        for (int j = u; j < newcrossings.length; j++) newcrossings[j] = crossings[j+1];
                        int[][] newpaths = new int[paths.length-1][4];
                        for (int j = 0; j < i; j++) {
                            for (int k = 0; k < 4; k++) 
                                newpaths[j][k] = theNewNumber(paths[j][k],s1,s2,2*paths.length);
                        }
                        newpaths[i][q1] = theNewNumber(paths[i][q1],s1,s2,2*paths.length);
                        newpaths[i][q2] = theNewNumber(paths[i][q2],s1,s2,2*paths.length);
                        newpaths[i][(q1+1)%4] = theNewNumber(paths[u][p1],s1,s2,2*paths.length);
                        newpaths[i][(q1+2)%4] = theNewNumber(paths[u][p2],s1,s2,2*paths.length);
                        for (int j = i+1; j < u; j++) {
                            for (int k = 0; k < 4; k++)
                                newpaths[j][k] = theNewNumber(paths[j][k],s1,s2,2*paths.length);
                        }
                        for (int j = u; j < newpaths.length; j++) {
                            for (int k = 0; k < 4; k++)
                                newpaths[j][k] = theNewNumber(paths[j+1][k],s1,s2,2*paths.length);
                        }
                        paths = newpaths;
                        crossings = newcrossings;
                        s1 = paths[i][(q1+1)%4];
                        s2 = paths[i][(q1+2)%4];
                    }
                    u++;
                }
                if (!found) {
                    sequence = false;
                    i++;
                }
            }
            else {
                q2 = 0;
                q1 = 1;
                boolean found = false;
                while (!found & q2 < 4) {
                    s1 = paths[i][(q1+1)%4];
                    s2 = paths[i][(q1+2)%4];
                    int u = i+1;
                    while (!found & u < paths.length) {
                        if (Math.abs(crossings[i]) == 1 || (q1%2 == 0)) {
                            if (contains(paths[u], s1, s2, crossings[u])) found = true;
                        }
                        if (!found) u++;
                    }
                    if (!found) {
                        q2++;
                        q1 = (q2+1)%4;
                    }
                }
                if (found) {
                    sequence = true;
                    if (q2%2 ==0) {
                        int help = paths[i][0];
                        paths[i][0] = paths[i][1];
                        paths[i][1] = paths[i][2];
                        paths[i][2] = paths[i][3];
                        paths[i][3] = help;
                        crossings[i] = - crossings[i];
                        q1 = (q1+3)%4;
                        q2 = (q2+3)%4;
                        for (int[] or : ors) {
                            if (or[0] == i) or[1] = (or[1]+3)%4;
                        }
                    }
                }
                else i++;
            }
        }
        return new Link(crossings, paths, ors, link.unComponents());
    }
    
    private static boolean contains(int[] path, int a, int b, int crs) {
        boolean found = false;
        int i = 0;
        int inc = 1;
        if (Math.abs(crs) > 1) inc++;
        while (!found && i < 4) {
            if (path[i] == a && path[(i+3)%4] == b) found = true;
            else i = i+inc;
        }
        return found;
    }
    
    private static int theNewNumber(int old, int s1, int s2, int max) {
        int nu = old;
        int big = s2;
        int sma = s1;
        if (s1 > s2) {
            big = s1;
            sma = s2;
        }
        if (sma == 1 || big == max) {
            if (sma == 1) {
                if (old < big) nu = old - 1;
                else nu = old -2;
            }
            else {
                if (old >= sma) nu = old - 1;
            }
        }
        else {
            if (old >= sma && old <= big) nu = old - 1;
            if (old > big) nu = old - 2;
        }
        return nu;
    }
    
}
