/*

Copyright (C) 2021 Dirk Schuetz <dirk.schuetz@durham.ac.uk>

This file is part of KnotJob.

KnotJob is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

KnotJob is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org.licenses/>.

 */

package knotjob.polynomial;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;

/**
 *
 * @author Dirk
 */
public class HalfPolynomial extends Polynomial {
    
    public HalfPolynomial(String[] lbls, BigInteger val, int[] coef) {
        super(lbls, new ArrayList<Coefficient>());
        coefficients.add(new HalfCoefficient(coef, val));
    }
    
    public HalfPolynomial(String[] lbls, ArrayList<Coefficient> coeffs) {
        super(lbls, coeffs);
    }
    
    public HalfPolynomial(String[] lbls, String reducedInfo) {
        super(lbls, new ArrayList<Coefficient>());
        if ("".equals(reducedInfo)) return;
        int firstComma = reducedInfo.indexOf(",");
        int q = Integer.parseInt(reducedInfo.substring(0, firstComma));
        int val;
        reducedInfo = reducedInfo.substring(firstComma+1);
        do {
            firstComma = reducedInfo.indexOf(",");
            if (firstComma >= 0) val = Integer.parseInt(reducedInfo.substring(0, firstComma));
            else val = Integer.parseInt(reducedInfo);
            if (val != 0) coefficients.add(new HalfCoefficient(new int[] {q}, BigInteger.valueOf(val)));
            q = q + 2;
            reducedInfo = reducedInfo.substring(firstComma+1);
        } while (firstComma >= 0);
    }
    
    public HalfPolynomial(Polynomial poly, String[] lbls, int[] div) {
        super(lbls, new ArrayList<Coefficient>());
        for (int i = 0; i < poly.getLength(); i++) {
            Coefficient coff = poly.getCoefficient(i);
            int[] pwrs = new int[div.length];
            for (int j = 0; j < div.length; j++) {
                pwrs[j] = coff.getPower(j)/div[j];
            }
            coefficients.add(new HalfCoefficient(pwrs, coff.getValue()));
        }
        Collections.sort(coefficients);
    }
    
    public HalfPolynomial multiply(HalfPolynomial pol) {
        ArrayList<Coefficient> mult = new ArrayList<Coefficient>();
        for (Coefficient coe : coefficients) {
            for (Coefficient col : pol.coefficients) {
                mult.add(((HalfCoefficient) coe).multiply((HalfCoefficient) col));
            }
        }
        Collections.sort(mult);
        combineCoefficients(mult);
        return new HalfPolynomial(labels, mult);
    }
    
    public HalfPolynomial add(HalfPolynomial pol) {
        ArrayList<Coefficient> add = new ArrayList<Coefficient>();
        for (Coefficient coe : coefficients) add.add((HalfCoefficient) coe);
        for (Coefficient coe : pol.coefficients) add.add((HalfCoefficient) coe);
        Collections.sort(add);
        combineCoefficients(add);
        return new HalfPolynomial(labels, add);
    }
    
    @Override
    protected void combineCoefficients(ArrayList<Coefficient> mult) {
        int i = mult.size()-1;
        while (i > 0) {
            HalfCoefficient coOne = (HalfCoefficient) mult.get(i-1);
            HalfCoefficient coTwo = (HalfCoefficient) mult.get(i);
            if (coOne.compareTo(coTwo) == 0) {
                HalfCoefficient comb = coOne.add(coTwo);
                mult.remove(i);
                mult.remove(i-1);
                if (comb.isZero()) i--;
                else mult.add(i-1,comb);
            }
            i--;
        }
    }
    
    public HalfPolynomial mirror() {
        ArrayList<Coefficient> mcoeffs = new ArrayList<Coefficient>();
        for (Coefficient coeff : coefficients) {
            int[] npowers = new int[coeff.powers.length];
            for (int i = 0; i < coeff.powers.length; i++) 
                npowers[i] = coeff.powers[i] * (-1);
            mcoeffs.add(new HalfCoefficient(npowers, coeff.getValue()));
        }
        Collections.sort(mcoeffs);
        return new HalfPolynomial(this.labels, mcoeffs);
    }
    
    public String toReducedString() {
        if (coefficients.isEmpty()) return "";
        int firstq = coefficients.get(0).getPower(0);
        String info = firstq+","+coefficients.get(0).value;
        for (int i = 1; i < coefficients.size(); i++) {
            firstq = firstq + 2;
            while (firstq != coefficients.get(i).getPower(0)) {
                firstq = firstq + 2;
                info = info + ",0";
            }
            info = info + ","+coefficients.get(i).value;
        }
        return info;
    }
    
    @Override
    public String toString() {
        String info = "";
        if (coefficients.isEmpty()) return "0";
        int i = 0;
        while (i < coefficients.size()-1) {
            info = info + coefficients.get(i).toString(latex, labels);
            if (coefficients.get(i+1).getValue().compareTo(BigInteger.ZERO) >= 0) info = info+" +"; 
            i++;
        }
        info = info + coefficients.get(coefficients.size()-1).toString(latex, labels);
        return info;
    }
}
